/*
 * Copyright (c) 2016-2020 Fairphone B.V.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#define LOG_TAG "Hiccupd"

#include "BnHiccupd.hpp"
#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <memory>
#include <string>

#include <android-base/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <binder/IPCThreadState.h>
#include <log/log.h>

using namespace android;
using namespace hiccup;

namespace {

enum class BootState {
  early,      // Early boot state -- hiccupd just started.
  hardReset,  // Hiccup app detected one boot event -- a hard reset.
  softReset,  // Hiccup app detected multiple boots -- a soft reset happened.
};

BootState boot_state = BootState::early;

void registerHiccupAppBootTrigger() {
  switch (boot_state) {
    case BootState::early:
      boot_state = BootState::hardReset;
      break;
    case BootState::hardReset:
      boot_state = BootState::softReset;
    case BootState::softReset:
      break;
  }
}

bool softResetOccurred() {
  return boot_state == BootState::softReset;
}

std::string copyLastCrashKernelLogIntoCache() {
  // For now use the same output name in both cases -- so that we don't need to
  // adjust legacy services to understand the new filename.
  static const std::string fileName = "/cache/hiccup/last_kmsg";
  int result = -1;
  if (softResetOccurred()) {
    /* Use dmesg to dump the current kernel log */
    result = system(("dmesg > " + fileName).c_str());
    ALOGI("Soft reset: Sending current kernel logs (dmesg).");
  } else {
    result = system(("cp /proc/last_kmsg " + fileName).c_str());
    ALOGI("Hard reset: Sending last kernel logs (/proc/last_kmsg).");
  }
  if (result != 0) {
    ALOGE("Failure while copying kernel logs to cache file %s: %i",
      fileName.c_str(), result);
  }
  return fileName;
}

}


class Hiccupd : public BnHiccupd {
  Vector<LogFile> *getLastKmsg();
  String16 getBootReason();
  String16 getPowerOnReason();
  String16 getPowerOffReason();
  int32_t  getApiVersion();
  int readIntFromFile(const char *fileName);
};

/** Get kernel logs potentially related to the last crash.
  *
  * On hard resets, we want the last kernel logs, because they potentially
  * contain relevant information about the reason of the crash. On soft resets,
  * only the Android framework, but not the Linux kernel reset, so we want the
  * current Linux kernel logs instead.
  */
Vector<LogFile> *Hiccupd::getLastKmsg() {
  /* make local copy of logs */
  system("rm -rf /cache/hiccup");
  system("mkdir -p /cache/hiccup");
  const std::string logFileName = copyLastCrashKernelLogIntoCache();

  auto v = std::make_unique<Vector<LogFile>>();
  const int fd = open(logFileName.c_str(), O_RDONLY);
  if (fd >= 0) {
    v->push_back(LogFile(String16(logFileName.c_str()), fd));
  } else {
    ALOGE("Failure opening the cached log file %s: Error %s (%i)",
      logFileName.c_str(), strerror(errno), errno);
  }

  system("rm -rf /cache/hiccup/");
  return v.release();
}

int Hiccupd::readIntFromFile(const char *fileName) {
  std::string content;
  if (!android::base::ReadFileToString(fileName, &content)) {
    ALOGE("Failure reading file %s", fileName);
    return -1;
  }
  return std::atoi(content.c_str());
}

String16 Hiccupd::getBootReason() {
  // NOTE: This assumes that the Hiccup app will only check the boot reason once
  // it detects a system boot, but never a second time while it is running.
  registerHiccupAppBootTrigger();
  if (softResetOccurred()) {
    return String16("soft reset");
  }
  String16 ret;
  int boot_reason = readIntFromFile("/proc/sys/kernel/boot_reason");
  switch (boot_reason) {
    case 0x01: ret = String16("keyboard power on"); break;
    case 0x02: ret = String16("RTC alarm"); break;
    case 0x04: ret = String16("cable power on"); break;
    case 0x08: ret = String16("SMPL"); break;
    case 0x10: ret = String16("Watch Dog timeout"); break;
    case 0x20: ret = String16("USB charger"); break;
    case 0x40: ret = String16("Wall charger"); break;
    case 0xFF: ret = String16("error reading power_on_status value"); break;
    default:   ret = String16(String8::format("Unknown: %d", boot_reason));

  }
  return ret;
}

String16 Hiccupd::getPowerOnReason() {
  String16 ret;
  int power_reason = readIntFromFile("/proc/sys/qpnp-power-on/pon_reason");
  switch (power_reason) {
    case 0:  ret = String16("(0) Triggered from Hard Reset"); break;
    case 1:  ret = String16("(1) Triggered from SMPL (sudden momentary power loss)"); break;
    case 2:  ret = String16("(2) Triggered from RTC (RTC alarm expiry)"); break;
    case 3:  ret = String16("(3) Triggered from DC (DC charger insertion)"); break;
    case 4:  ret = String16("(4) Triggered from USB (USB charger insertion)"); break;
    case 5:  ret = String16("(5) Triggered from PON1 (secondary PMIC)"); break;
    case 6:  ret = String16("(6) Triggered from CBL (external power supply)"); break;
    case 7:  ret = String16("(7) Triggered from KPD (power key press)"); break;
    default: ret = String16(String8::format("(%d) Unknown", power_reason));
  }
  return ret;
}

String16 Hiccupd::getPowerOffReason() {
  String16 ret;
  int power_reason = readIntFromFile("/proc/sys/qpnp-power-on/poff_reason");
  switch (power_reason) {
    case 0:  ret = String16("(0) Triggered from SOFT (Software)"); break;
    case 1:  ret = String16("(1) Triggered from PS_HOLD (PS_HOLD/MSM controlled shutdown)"); break;
    case 2:  ret = String16("(2) Triggered from PMIC_WD (PMIC watchdog)"); break;
    case 3:  ret = String16("(3) Triggered from GP1 (Keypad_Reset1)"); break;
    case 4:  ret = String16("(4) Triggered from GP2 (Keypad_Reset2)"); break;
    case 5:  ret = String16("(5) Triggered from KPDPWR_AND_RESIN (Simultaneous power key and reset line)"); break;
    case 6:  ret = String16("(6) Triggered from RESIN_N (Reset line/Volume Down Key)"); break;
    case 7:  ret = String16("(7) Triggered from KPDPWR_N (Long Power Key hold)"); break;
    case 8:  ret = String16("(8) N/A 8"); break;
    case 9:  ret = String16("(9) N/A 9"); break;
    case 10: ret = String16("(10) N/A 10"); break;
    case 11: ret = String16("(11) Triggered from CHARGER (Charger ENUM_TIMER, BOOT_DONE)"); break;
    case 12: ret = String16("(12) Triggered from TFT (Thermal Fault Tolerance)"); break;
    case 13: ret = String16("(13) Triggered from UVLO (Under Voltage Lock Out)"); break;
    case 14: ret = String16("(14) Triggered from OTST3 (Overtemp)"); break;
    case 15: ret = String16("(15) Triggered from STAGE3 (Stage 3 reset)"); break;
    default: ret = String16(String8::format("(%d) Unknown", power_reason));
  }
  return ret;
}

int32_t Hiccupd::getApiVersion() {
  return HICCUP_API_VERSION;
}

int main() {
  defaultServiceManager()->addService(String16("com.fairphone.hiccupd"), new Hiccupd());
  ProcessState::self()->startThreadPool();
  IPCThreadState::self()->joinThreadPool();
  return 0;
}
